#include "app.hh"

App::App(const char *title, int w, int h, float scale):
	scrW(w / scale),
	scrH(h / scale),
	scrSize(scrW * scrH),
	pixels(new std::uint32_t[scrSize]),

	winW(w),
	winH(h),
	aspectRatio((float)h / w),

	running(true),
	world(10, 10, Camera(Vec2f(3, 8), M_PI * 1.75, 70, 0.2))
{
	// SDL2 setup
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		std::cerr << "Failed to initialize SDL2: " << SDL_GetError() << std::endl;
		std::exit(EXIT_FAILURE);
	}

	win = SDL_CreateWindow(title, SDL_WINDOWPOS_CENTERED, SDL_WINDOWPOS_CENTERED,
	                       w, h, SDL_WINDOW_SHOWN);
	if (win == NULL) {
		std::cerr << "Failed to create window: " << SDL_GetError() << std::endl;
		std::exit(EXIT_FAILURE);
	}

	ren = SDL_CreateRenderer(win, -1, SDL_RENDERER_ACCELERATED | SDL_RENDERER_PRESENTVSYNC);
	if (ren == NULL) {
		std::cerr << "Failed to create renderer: " << SDL_GetError() << std::endl;
		std::exit(EXIT_FAILURE);
	}

	tex = SDL_CreateTexture(ren, SDL_PIXELFORMAT_RGBA8888, SDL_TEXTUREACCESS_STREAMING,
	                        scrW, scrH);
	if (tex == NULL) {
		std::cerr << "Failed to create screen texture: " << SDL_GetError() << std::endl;
		std::exit(EXIT_FAILURE);
	}

	keyboard = SDL_GetKeyboardState(NULL);

	// A hack to force i3wm/sway to float the window on startup
	SDL_SetWindowResizable(win, SDL_TRUE);

	// Init x dir map
	xDirMap.resize(scrW);

	double centerX   = (double)(scrW / 2);
	double planeDist = centerX / std::tan(degToRad(world.cam.fov / 2));
	for (int x = 0; x < scrW; ++ x)
		xDirMap[x] = std::atan2((double)x - centerX, planeDist);

	// Setup map
	world.addLines({
		Linef(2, 3, 3, 2),
		Linef(2, 3, 2, 4),
		Linef(2, 3, 1.8, 3.5),
		Linef(2, 4, 1.8, 3.5),
		Linef(4, 2, 3, 2),
		Linef(4, 2, 5, 3),
		Linef(5, 3, 5, 4),
		Linef(5, 4, 4, 5),
		Linef(4, 5, 3, 5),
		Linef(3, 5, 3.5, 5.2),
		Linef(4, 5, 3.5, 5.2),

		Linef(7, 7, 7, 6),
		Linef(6.5, 7, 6.5, 6),
		Linef(6.5, 7, 7, 7),
		Linef(6.5, 6, 7, 6),

		Linef(6.5, 6, 7.8, 8),
		Linef(6.5, 6, 7.4, 7),
		Linef(7.8, 8, 7.4, 7),

		Linef(0, 6, 2, 9),
		Linef(0, 5, 2, 9),

		Linef(8, 2, 7.2, 3.2),
		Linef(6, 4, 7.2, 3.2),
		Linef(8, 2, 6.8, 2.8),
		Linef(6, 4, 6.8, 2.8),

		Linef(8, 4, 7.2, 2.8),
		Linef(6, 2, 7.2, 2.8),
		Linef(8, 4, 6.8, 3.2),
		Linef(6, 2, 6.8, 3.2),
	});
}

App::~App() {
	free(pixels);
	SDL_DestroyTexture(tex);
	SDL_DestroyRenderer(ren);
	SDL_DestroyWindow(win);

	SDL_Quit();
}

void App::run() {
	uint64_t last, now = SDL_GetPerformanceCounter();
	while (running) {
		last = now;
		now  = SDL_GetPerformanceCounter();
		double dt = (double)(now - last) * 1000 / (double)SDL_GetPerformanceFrequency();

		// Render
		SDL_SetRenderDrawColor(ren, 0, 0, 0, SDL_ALPHA_OPAQUE);
		SDL_RenderClear(ren);

		std::memset(pixels, 0, scrSize * sizeof(*pixels));
		render(dt);
		SDL_UpdateTexture(tex, NULL, pixels, scrW * sizeof(*pixels));

		// Display
		float winAspectRatio = (float)winH / winW;
		viewport.w = winAspectRatio < aspectRatio? winH / aspectRatio : winW;
		viewport.h = winAspectRatio > aspectRatio? winW * aspectRatio : winH;
		viewport.x = winW / 2 - viewport.w / 2;
		viewport.y = winH / 2 - viewport.h / 2;
		SDL_RenderCopy(ren, tex, NULL, &viewport);
		SDL_RenderPresent(ren);

		// Handle events
		SDL_Event evt;
		while (SDL_PollEvent(&evt))
			handleEvent(evt);

		update(dt);
	}
}

void App::render(double dt) {
	(void)dt;

	double fog = 14;

	for (int x = 0; x < scrW; ++ x) {
		RaycastResult result = world.raycast(xDirMap[x]);

		double dirCos = std::cos(xDirMap[x]);
		double dist   = distance(world.cam.pos, result.hit);
		int    wallH  = std::round((double)scrH / (dist * dirCos));
		int    wallY  = scrH / 2 - wallH / 2;

		if (wallH > scrH) wallH = scrH;
		if (wallY < 0)    wallY = 0;

		// Render wall
		float t     = std::max(1.0 - (result.shade * 3 + dist) / fog, 0.0);
		auto  color = (std::uint32_t)(t * 255);
		for (int yo = 0; yo < wallH; ++ yo)
			pixels[(yo + wallY) * scrW + x] = color << 24;

		// Render floor
		for (int yo = wallH / 2; yo < scrH / 2; ++ yo) {
			double dist = (double)scrH / 2 / yo / dirCos;
			Vec2i p(world.cam.pos + Vec2f::fromDir(world.cam.dir + xDirMap[x]) * dist);

			float t     = std::max(1.0 - dist / fog, 0.0);
			auto  color = (std::uint32_t)(t * (p.x % 2 == (p.y % 2 == 0)? 170 : 70));

			pixels[(yo + scrH / 2) * scrW + x] = color << 24 | color << 16 | color << 8;
		}
	}
}

template<typename T>
bool lineIntersectsRect(const Line<T> &l, const Vec2<T> &p1, const Vec2<T> &p2, Vec2<T> &i) {
	std::array<Line<T>, 4> sides = {
		Line<T>(p1.x, p1.y, p1.x, p2.y),
		Line<T>(p1.x, p1.y, p2.x, p1.y),
		Line<T>(p2.x, p2.y, p1.x, p2.y),
		Line<T>(p2.x, p2.y, p2.x, p1.y),
	};

	for (const auto &side : sides) {
		if (l.intersects(side, i))
			return true;
	}
	return false;
}

void App::update(double dt) {
	Vec2f  prev   = world.cam.pos;
	double moveBy = 0.005 * dt;

	if (keyboard[SDL_SCANCODE_W]) world.cam.moveForward(moveBy);
	if (keyboard[SDL_SCANCODE_S]) world.cam.moveBackward(moveBy);
	if (keyboard[SDL_SCANCODE_A]) world.cam.moveLeft(moveBy);
	if (keyboard[SDL_SCANCODE_D]) world.cam.moveRight(moveBy);
	if (keyboard[SDL_SCANCODE_Q]) world.cam.turnLeft(moveBy);
	if (keyboard[SDL_SCANCODE_E]) world.cam.turnRight(moveBy);

	Vec2f p1 = world.cam.pos - world.cam.radius;
	Vec2f p2 = world.cam.pos + world.cam.radius;

	// Collision with lines
	for (const auto &l : world.lines) {
		Vec2f i;
		if (lineIntersectsRect(l, p1, p2, i)) {
			world.cam.pos = prev;
			break;
		}
	}

	// Collision with borders
	if (p1.x < 0)       world.cam.pos.x = world.cam.radius;
	if (p1.y < 0)       world.cam.pos.y = world.cam.radius;
	if (p2.x > world.w) world.cam.pos.x = world.w - world.cam.radius;
	if (p2.y > world.h) world.cam.pos.y = world.h - world.cam.radius;
}

void App::handleEvent(SDL_Event &evt) {
	switch (evt.type) {
	case SDL_QUIT: running = false; break;
	case SDL_WINDOWEVENT:
		if (evt.window.event == SDL_WINDOWEVENT_RESIZED) {
			winW = evt.window.data1;
			winH = evt.window.data2;
		}
		break;
	}
}
