#ifndef MATH_HH_HEADER_GUARD
#define MATH_HH_HEADER_GUARD

#include <cmath>     // std::sqrt, std::sqrt, std::round, std::cos, std::sin, std::pow
#include <algorithm> // std::swap
#include <array>     // std::array

#ifndef M_PI
#	define M_PI 3.1415926535
#endif

constexpr double lineBoundaryOffset = 0.00005;

template<typename T>
struct Vec2 {
	T x, y;

	static Vec2<T> fromDir(T dir) {
		return Vec2<T>(std::cos(dir), std::sin(dir));
	}

	Vec2():         x(0), y(0) {}
	Vec2(T a):      x(a), y(a) {}
	Vec2(T a, T b): x(a), y(b) {}

#define funcTemplate(OP)                    \
	Vec2 operator OP(const Vec2 &p) const { \
		return Vec2<T>(x OP p.x, y OP p.y); \
	}

	funcTemplate(+)
	funcTemplate(-)
	funcTemplate(*)
	funcTemplate(/)

#undef  funcTemplate
#define funcTemplate(OP)                \
	Vec2 &operator OP (const Vec2 &p) { \
		x OP p.x;                       \
		y OP p.y;                       \
		return *this;                   \
	}

	funcTemplate(+=)
	funcTemplate(-=)
	funcTemplate(*=)
	funcTemplate(/=)

#undef  funcTemplate
#define funcTemplate(T2)               \
	operator Vec2<T2>() const {        \
		return Vec2<T2>((T2)x, (T2)y); \
	}

	funcTemplate(float)
	funcTemplate(int)

#undef  funcTemplate
#define funcTemplate(NAME, FUNC)          \
	Vec2<T> NAME() const {                \
		return Vec2<T>(FUNC(x), FUNC(y)); \
	}

	funcTemplate(round, std::round)
	funcTemplate(ceil,  std::ceil)
	funcTemplate(floor, std::floor)

#undef funcTemplate

	bool operator ==(const Vec2 &p) {
		return x == p.x and y == p.y;
	}

	bool operator !=(const Vec2 &p) {
		return x != p.x or y != p.y;
	}

	Vec2<T> swap() const {
		return Vec2(y, x);
	}

	Vec2<T> round(double precision) const {
		precision = std::pow(10, precision);
		return Vec2<T>(std::round(x * precision) / precision,
		               std::round(y * precision) / precision);
	}

	friend std::ostream &operator <<(std::ostream &stream, const Vec2 &p) {
		stream << '{' << p.x << ", " << p.y << '}';
	    return stream;
	}
};

using Vec2f = Vec2<double>;
using Vec2i = Vec2<int>;

template<typename T>
struct Line {
	Vec2<T> a, b;

	Line() {}
	Line(Vec2<T> a, Vec2<T> b):   a(a),      b(b) {}
	Line(T x1, T y1, T x2, T y2): a(x1, y1), b(x2, y2) {}

	bool isPointBetweenVertices(const Vec2<T> &p, double boundary) const {
		Vec2<T> p1 = a, p2 = b;

		if (p1.x > p2.x)
			std::swap(p1.x, p2.x);
		if (p.x > p2.x + boundary || p.x < p1.x - boundary)
			return false;

		if (p1.y > p2.y)
			std::swap(p1.y, p2.y);
		if (p.y > p2.y + boundary || p.y < p1.y - boundary)
			return false;

		return true;
	}

	bool intersects(const Line<T> &l, Vec2<T> &i) const {
		double a1 = b.y - a.y;
		double b1 = a.x - b.x;
		double c1 = a1 * a.x + b1 * a.y;

		double a2 = l.b.y - l.a.y;
		double b2 = l.a.x - l.b.x;
		double c2 = a2 * l.a.x + b2 * l.a.y;

		double determinant = a1 * b2 - a2 * b1;
		if (determinant == 0)
			return false;

		i.x = (b2 * c1 - b1 * c2) / determinant;
		i.y = (a1 * c2 - a2 * c1) / determinant;
		return isPointBetweenVertices(i, lineBoundaryOffset) &&
		       l.isPointBetweenVertices(i, lineBoundaryOffset);
	}
};

using Linef = Line<double>;

template<typename T>
double distance(const Vec2<T> &a, const Vec2<T> &b) {
	return std::sqrt(std::pow(a.x - b.x, 2) + std::pow(a.y - b.y, 2));
}

inline double degToRad(double deg) {
	return deg * M_PI / 180;
}

inline double lerp(float a, float b, float f) {
	return a + f * (b - a);
}

#endif
