#include "world.hh"

Tile::Tile() {}

Camera::Camera(Vec2f pos, double dir, double fov, double radius):
	pos(pos),
	dir(dir),
	fov(fov),
	radius(radius)
{}

void Camera::moveForward(double scale) {
	pos += Vec2f::fromDir(dir) * scale;
}

void Camera::moveBackward(double scale) {
	pos -= Vec2f::fromDir(dir) * scale;
}

void Camera::moveLeft(double scale) {
	pos += Vec2f::fromDir(dir).swap() * Vec2f(scale, -scale);
}

void Camera::moveRight(double scale) {
	pos -= Vec2f::fromDir(dir).swap() * Vec2f(scale, -scale);
}

void Camera::turnLeft(double scale) {
	dir -= scale;
	sanitizeDir();
}

void Camera::turnRight(double scale) {
	dir += scale;
	sanitizeDir();
}

void Camera::sanitizeDir() {
	while (dir > M_PI * 2) dir -= M_PI * 2;
	while (dir < 0)        dir += M_PI * 2;
}

World::World(int w, int h, const Camera &cam):
	w(w),
	h(h),

	cam(cam)
{
	tiles.resize(h);
	for (auto &row : tiles)
		row.resize(w);
}

const Tile &World::tileAt(int x, int y) const {
	return tiles[y][x];
}

void World::initDDA(Vec2f start, Vec2f d, Vec2f &m,
                    Vec2f &l, Vec2f &ul, Vec2i &p, Vec2i &step) const {
	m  = Vec2f(d.x / d.y, d.y / d.x);
	ul = Vec2f(std::hypot(1, m.y), std::hypot(m.x, 1));
	p  = Vec2i(start);

	if (d.x < 0) {
		step.x = -1;
		l.x    = (start.x - p.x) * ul.x;
	} else {
		step.x = 1;
		l.x    = (p.x + 1 - start.x) * ul.x;
	}

	if (d.y < 0) {
		step.y = -1;
		l.y    = (start.y - p.y) * ul.y;
	} else {
		step.y = 1;
		l.y    = (p.y + 1 - start.y) * ul.y;
	}
}

void World::findClosestLineIntersect(const Vec2i &p, const Linef &line,
                                     Vec2f &closestIntersect, const Linef *&closestLine) const {
	closestLine = NULL;

	double closestDist = -1;
	for (const auto *l : tiles[p.y][p.x].lines) {
		Vec2f i;
		bool intersects = line.intersects(*l, i);
		if (intersects &&
		    i.x >= p.x - lineBoundaryOffset     && i.y >= p.y - lineBoundaryOffset &&
		    i.x <  p.x + lineBoundaryOffset + 1 && i.y <  p.y + lineBoundaryOffset + 1) {
			double dist = distance(line.a, i);
			if (dist < closestDist || closestDist < 0) {
				closestDist      = dist;
				closestIntersect = i;
				closestLine      = l;
			}
		}
	}
}

RaycastResult World::raycast(double off) const {
	Vec2f d(Vec2f::fromDir(cam.dir + off));

	Vec2f m, l, ul;
	Vec2i p, step;
	initDDA(cam.pos, d, m, l, ul, p, step);

	double dist = 0;
	RaycastResult result;
	while (true) {
		if (!tiles[p.y][p.x].lines.empty()) {
			Linef ray(cam.pos, cam.pos + d * (dist + 2));

			const Linef *line;
			findClosestLineIntersect(p, ray, result.hit, line);
			if (line != NULL) {
				result.shade = std::fabs(line->a.y - line->b.y) / distance(line->a, line->b);
				return result;
			}
		}

		if (l.x < l.y) {
			dist = l.x;
			p.x += step.x;
			l.x += ul.x;

			result.shade = 1;
		} else {
			dist = l.y;
			p.y += step.y;
			l.y += ul.y;

			result.shade = 0;
		}

		if (p.x >= w || p.y >= h || p.x < 0 || p.y < 0)
			break;
	}

	result.hit = cam.pos + d * dist;
	return result;
}

void World::addLines(const std::vector<Linef> &toAdd) {
	lines.insert(lines.end(), toAdd.begin(), toAdd.end());
	mapLinesToTiles();
}

void World::mapLineToTiles(const Linef &line) {
	Vec2f a = line.a, b = line.b;
	if (a.y > b.y)
		std::swap(a, b);

	double dist = distance(a, b);

	Vec2f o(b.x - a.x, b.y - a.y);

	Vec2f m, l, ul;
	Vec2i p, step;
	initDDA(a, o, m, l, ul, p, step);

	bool skip = false;
	while (true) {
		auto &lines = tiles[p.y][p.x].lines;
		if (!skip && std::find(lines.begin(), lines.end(), &line) == lines.end())
			lines.push_back(&line);

		skip = l.x == l.y;

		if (l.x < l.y) {
			if (l.x > dist)
				break;

			p.x += step.x;
			l.x += ul.x;
		} else {
			if (l.y > dist)
				break;

			p.y += step.y;
			l.y += ul.y;
		}

		if (p.x >= w || p.y >= h || p.x < 0 || p.y < 0)
			break;
	}
}

void World::mapLinesToTiles() {
	for (auto &row : tiles) {
		for (auto &tile : row)
			tile.lines.clear();
	}

	for (const auto &line : lines)
		mapLineToTiles(line);
}
